from typing import Optional, List, NamedTuple, Union
import wget
import shutil
from pathlib import Path
import os
import subprocess
import torch
from rich import print
import pytorch_lightning as pl
import argparse
from clipzyme.models.protmol import EnzymeReactionCLIP
from clipzyme.utils.screening import process_mapped_reaction
from clipzyme.utils.protein_utils import create_protein_graph
from clipzyme.utils.loading import default_collate

CHECKPOINT_URL = "https://zenodo.org/records/11187895/files/clipzyme_model.zip"


def download_and_extract(remote_model_url: str, local_model_dir: str) -> str:
    """
    Download and extract model.

    Parameters
    ----------
    remote_model_url : str
        link to model file
    local_model_dir : str
        local directory to save model

    Returns
    -------
    str
        name of model file
    """
    zipfile = wget.download(remote_model_url, out=local_model_dir)
    # unzip
    subprocess.call(f"unzip {zipfile} -d {local_model_dir}", shell=True)
    # remove zip
    os.remove(zipfile)
    # get all files and dirs
    all_files_and_dirs = os.listdir(local_model_dir)
    assert len(all_files_and_dirs) == 1, "More than one file in model directory"
    return all_files_and_dirs[0]


def download_model(checkpoint_path) -> str:
    """
    Download trained clipzyme model.

    Parameters
    ----------
        checkpoint_path (str): path to where model should be located

    Returns
    -------
        str: local path to model
    """
    # Create cache folder if not exists
    cache = os.path.dirname(checkpoint_path)
    os.makedirs(cache, exist_ok=True)

    # Download model
    print(f"Downloading model to {cache}")
    filename = download_and_extract(CHECKPOINT_URL, cache)
    model_path = os.path.join(cache, filename)
    print(f"Model saved in {model_path}")
    return model_path


class CLIPZymeOutput(NamedTuple):
    scores: torch.Tensor
    protein_hiddens: torch.Tensor
    reaction_hiddens: torch.Tensor
    sample_ids: List[str]


class CLIPZyme(pl.LightningModule):
    def __init__(
        self,
        args: argparse.Namespace = None,
        checkpoint_path: str = None,
        device: Optional[str] = None,
    ) -> None:
        """
        Initialize a trained clipzyme model for inference.

        Parameters
        ----------
        args: argparse.Namespace
            Arguments from command line.
        checkpoint_path: str
            Path to a clipzyme checkpoint.
        device: str
            If provided, will run inference using this device.
            By default uses GPU, if available.
        """
        super(CLIPZyme, self).__init__()
        if args is not None:
            checkpoint_path = args.checkpoint_path
            self.use_as_protein_encoder = args.use_as_protein_encoder
            self.use_as_reaction_encoder = args.use_as_reaction_encoder

        # Check if path exists
        if (checkpoint_path is None) or (not os.path.exists(checkpoint_path)):
            try:
                # download model
                print(f"Model checkpoint not found at {checkpoint_path}")
                checkpoint_path = download_model(checkpoint_path)
            except Exception as e:
                raise FileNotFoundError(
                    f"Model checkpoint not found at {checkpoint_path} and failed to download model. {e}"
                )

        # Load model
        self.model = self.load_model(checkpoint_path, args)

        # prep directories
        self.prepare_directories(args)

        # Set device
        if device is not None:
            self.device = device
            self.model = self.model.to(self.device)

    def prepare_directories(self, args: argparse.Namespace) -> None:
        """
        Prepare directories for saving hiddens and predictions.

        Parameters
        ----------
        args : argparse.Namespace
            Arguments from command line.
        """
        self.save_hiddens = getattr(args, "save_hiddens", False)
        self.save_predictions = getattr(args, "save_predictions", False)
        if self.save_hiddens or self.save_predictions:
            checkpoint_path = Path(args.checkpoint_path)
            dataset_path = Path(args.dataset_file_path)
            hiddens_dir = os.path.join(
                args.inference_dir, checkpoint_path.stem, dataset_path.stem
            )
            os.makedirs(hiddens_dir, exist_ok=True)
            self.hiddens_dir = Path(hiddens_dir)

        if self.save_predictions:
            self.predictions_dir = os.path.join(hiddens_dir, "predictions")
            os.makedirs(self.predictions_dir, exist_ok=True)
            self.predictions_dir = Path(self.predictions_dir)
            self.predictions_path = os.path.join(self.hiddens_dir, "predictions.csv")

    def load_model(
        self, path: str, screen_args: argparse.Namespace
    ) -> EnzymeReactionCLIP:
        """
        Load model from path.

        Parameters
        ----------
        path : str
            Path to a clipzyme model checkpoint.

        Returns
        -------
        model
            Pretrained clipzyme model
        """
        # Load checkpoint
        checkpoint = torch.load(path, map_location="cpu")
        args = checkpoint["hyper_parameters"]["args"]
        # set relevant args
        for key in ["use_as_protein_encoder", "use_as_reaction_encoder"]:
            setattr(args, key, getattr(screen_args, key, False))
            setattr(self, key, getattr(screen_args, key, False))
        model = EnzymeReactionCLIP(args)

        # Remove model from param names
        state_dict = {k[6:]: v for k, v in checkpoint["state_dict"].items()}
        model.load_state_dict(state_dict)  # type: ignore

        # Set eval
        model.eval()
        print(f"[bold] Loaded model from {path}")
        return model

    def forward(
        self,
        batch,
        batch_idx: int = 0,
    ) -> CLIPZymeOutput:
        """
        Forward pass of the model.

        Parameters
        ----------
        batch : dict
            Batch of data.
        batch_idx : int, optional
            batch id, by default 0

        Returns
        -------
        CLIPZymeOutput
            Output of the model.
        """
        reaction_scores = None
        protein_hiddens = None
        reaction_hiddens = None

        model_output = self.model(batch)

        predictions_dict = {}
        predictions_dict = self.store_in_predictions(predictions_dict, batch)
        predictions_dict = self.store_in_predictions(predictions_dict, model_output)

        # want paired scores
        if not (self.use_as_protein_encoder or self.use_as_reaction_encoder):
            reaction_hiddens = model_output["substrate_hiddens"]
            protein_hiddens = model_output["protein_hiddens"]
            reaction_scores = torch.einsum(
                "bx,bx->b", reaction_hiddens, protein_hiddens
            )

        if self.use_as_protein_encoder:
            protein_hiddens = self.extract_protein_features(batch)
        if self.use_as_reaction_encoder:
            reaction_hiddens = self.extract_reaction_features(batch)

        output = CLIPZymeOutput(
            scores=reaction_scores,
            protein_hiddens=protein_hiddens,
            reaction_hiddens=reaction_hiddens,
            sample_ids=batch["sample_id"],
        )

        return output

    def test_step(self, batch, batch_idx) -> CLIPZymeOutput:
        """
        Test step for the model.

        Parameters
        ----------
        batch : dict
            Batch of data.
        batch_idx : int
            Batch index.

        Returns
        -------
        ClipZymeOutput
            Output of the model.
        """
        output = self.forward(batch, batch_idx)
        if self.save_hiddens or self.save_predictions:
            self.save_outputs(output)
        return

    def on_test_epoch_end(self) -> None:
        self.clean_up()

    def save_outputs(self, outputs: CLIPZymeOutput) -> None:
        """
        Save outputs to disk.

        Parameters
        ----------
        outputs : ClipZymeOutput
            Output of the model.
        """
        if self.save_hiddens:
            if outputs.protein_hiddens is not None:
                protein_hiddens = outputs.protein_hiddens.cpu()
                for idx, protein_hidden in enumerate(protein_hiddens):
                    predictions_filename = self.hiddens_dir.joinpath(
                        f"sample_{outputs.sample_ids[idx]}.protein.pt"
                    )
                    torch.save(protein_hidden, predictions_filename)

            if outputs.reaction_hiddens is not None:
                reaction_hiddens = outputs.reaction_hiddens.cpu()
                for idx, reaction_hidden in enumerate(reaction_hiddens):
                    predictions_filename = self.hiddens_dir.joinpath(
                        f"sample_{outputs.sample_ids[idx]}.reaction.pt"
                    )
                    torch.save(reaction_hidden, predictions_filename)

        if self.save_predictions:
            scores = outputs.scores.cpu()
            for idx, score in enumerate(scores):
                predictions_filename = self.predictions_dir.joinpath(
                    f"sample_{outputs.sample_ids[idx]}.score.pt"
                )
                torch.save(score.item(), predictions_filename)

    def extract_protein_features(
        self,
        batch: dict = None,
        cif_path: Union[str, List[str]] = None,
        esm_dir: str = None,
    ) -> torch.Tensor:
        """
        Extract protein features from model.

        Parameters
        ----------
        batch : dict
            Batch of data.
        cif_path : Union[str, List[str]], optional
            Path to CIF file, by default None
        esm_dir : str, optional
            Path to ESM model (esm2_t33_650M_UR50D), by default None

        Returns
        -------
        torch.Tensor
            Protein features.
        """
        if cif_path is not None:
            assert (
                esm_dir is not None
            ), "If manually extracting protein embedding, then `esm_dir` must be provided"
            if isinstance(cif_path, str):
                cif_path = [cif_path]
            protein_graphs = [
                create_protein_graph(
                    cif_path=cpath,
                    esm_path=os.path.join(esm_dir, "esm2_t33_650M_UR50D.pt"),
                )
                for cpath in cif_path
            ]
            batch = default_collate([{"graph": g} for g in protein_graphs])

        protein_features = self.model.encode_protein(batch)
        protein_features = protein_features / protein_features.norm(dim=1, keepdim=True)

        return protein_features

    def extract_reaction_features(
        self, batch: dict = None, reaction: Union[str, List[str]] = None
    ) -> torch.Tensor:
        """
        Extract reaction features from model.

        Parameters
        ----------
        batch : dict
            Batch of data.
        reaction : Union[str, List[str]], optional
            Mapped raction string, by default None

        Returns
        -------
        torch.Tensor
            Reaction features.
        """
        if reaction is not None:
            if isinstance(reaction, str):
                reaction = [reaction]
            reactions = [process_mapped_reaction(rxn) for rxn in reaction]
            batch = default_collate(
                [
                    {
                        "reactants": rxn[0],
                        "products": rxn[1],
                    }
                    for rxn in reactions
                ]
            )
        substrate_features = self.model.encode_reaction(batch)
        substrate_features = substrate_features / substrate_features.norm(
            dim=1, keepdim=True
        )

        return substrate_features

    def store_in_predictions(self, preds: dict, storage_dict: dict) -> dict:
        """
        Store values in predictions.

        Parameters
        ----------
        preds : dict
            prediction dictionary
        storage_dict : dict
            dictionary from which to get values

        Returns
        -------
        dict
            updated predictions dictionary
        """
        for key, val in storage_dict.items():
            if torch.is_tensor(val) and val.requires_grad:
                preds[key] = val.detach()
            else:
                preds[key] = val
        return preds

    def clean_up(self):
        """
        Clean up directories after inference.
        """
        if self.save_predictions:
            print("Collecting predictions files")

            with open(self.predictions_path, "w") as f:
                f.write("sample_id,score\n")
                for score_file in self.predictions_dir.iterdir():
                    score = torch.load(score_file)
                    f.write(f"{score_file.stem.split('.')[0]},{score}\n")

            shutil.rmtree(self.predictions_dir)
        print("Done")



import json, os
from typing import Union, NamedTuple
from pathlib import Path
import torch
import pickle
from p_tqdm import p_map
from functools import partial
import pandas as pd
from clipzyme.utils.wln_processing import get_bond_changes
from clipzyme.utils.pyg import from_mapped_smiles


class ScreeningResults(NamedTuple):
    predictions: pd.DataFrame
    protein_hiddens: torch.Tensor
    reaction_hiddens: torch.Tensor
    protein_ids: list
    reaction_ids: list


def load_hidden(hiddens_dir: Path, file: str) -> Union[torch.Tensor, None]:
    """
    Load hidden representation from a file

    Parameters
    ----------
    hiddens_dir : Path
        directory where hidden representations are saved
    file : str
        file name of hidden representation

    Returns
    -------
    Union[torch.Tensor, None]
        hidden representation tensor or None if file does not exist
    """
    try:
        return torch.load(hiddens_dir.joinpath(file))
    except FileNotFoundError:
        return


def collect_screening_results(config_path: str) -> ScreeningResults:
    """
    Collect screening results from a screening experiment

    Parameters
    ----------
    config_path : str
        path to screening experiment configuration file

    Returns
    -------
    ScreeningResults
        screening results, including predictions, protein hiddens, reaction hiddens, protein ids, and reaction ids
    """
    config = json.load(open(config_path, "r"))
    inference_dir = config["inference_dir"]
    save_hiddens = config["save_hiddens"]
    save_predictions = config["save_predictions"]
    checkpoint_path = Path(config["checkpoint_path"])
    dataset_path = Path(config["dataset_file_path"])
    # get directory where hiddens are saved
    hiddens_dir = os.path.join(inference_dir, checkpoint_path.stem, dataset_path.stem)

    predictions = None
    if save_predictions:
        # get path to predictions (screening scores)
        predictions_path = os.path.join(hiddens_dir, "predictions.csv")
        predictions = pd.read_csv(predictions_path)

    hiddens_dir = Path(hiddens_dir)
    assert hiddens_dir.is_dir(), f"{hiddens_dir} is not a directory"

    protein_hiddens, reaction_hiddens = None, None
    protein_ids, reaction_ids = None, None
    if save_hiddens:
        load_hidden_func = partial(load_hidden, path=hiddens_dir)
        protein_files = list(hiddens_dir.glob("sample_*.protein.pt"))
        reaction_files = list(hiddens_dir.glob("sample_*.reaction.pt"))

        load_hidden_func = partial(load_hidden, path=hiddens_dir)
        protein_hiddens = p_map(load_hidden_func, protein_files)
        reaction_hiddens = p_map(load_hidden_func, reaction_files)

        protein_ids = [f.stem.split("_")[-1] for f in protein_files]
        reaction_ids = [f.stem.split("_")[-1] for f in reaction_files]

        protein_ids = [u for u, h in zip(protein_ids, protein_hiddens) if h is not None]
        reaction_ids = [
            u for u, h in zip(reaction_ids, reaction_hiddens) if h is not None
        ]

        protein_hiddens = [h for h in protein_hiddens if h is not None]
        reaction_hiddens = [h for h in reaction_hiddens if h is not None]

        protein_hiddens = torch.stack(protein_hiddens)
        reaction_hiddens = torch.stack(reaction_hiddens)

    # collect
    output = ScreeningResults(
        predictions=predictions,
        protein_hiddens=protein_hiddens,
        reaction_hiddens=reaction_hiddens,
        protein_ids=protein_ids,
        reaction_ids=reaction_ids,
    )

    return output


def process_mapped_reaction(
    reaction: str, bond_changes=None, use_one_hot_mol_features: bool = False
):
    """
    Process a mapped reaction string into a PyG data object

    Parameters
    ----------
    reaction : str
        mapped reaction string
    bond_changes : list, optional
        list of bond change t (i,j,t) associated with edges (i,j), by default None
    use_one_hot_mol_features : bool, optional
        whether to use one-hot features for molecules , by default False

    Returns
    -------
    tuple of PyG data objects
        reactants and products as PyG data objects
    """
    reactants, products = reaction.split(">>")

    reactants, atom_map2new_index = from_mapped_smiles(
        reactants,
        encode_no_edge=True,
        use_one_hot_encoding=use_one_hot_mol_features,
    )
    products, _ = from_mapped_smiles(
        products,
        encode_no_edge=True,
        use_one_hot_encoding=use_one_hot_mol_features,
    )

    if bond_changes is None:
        bond_changes = get_bond_changes(reaction)

    bond_changes = [
        (atom_map2new_index[int(u)], atom_map2new_index[int(v)], btype)
        for u, v, btype in bond_changes
    ]
    bond_changes = [(min(x, y), max(x, y), t) for x, y, t in bond_changes]
    reactants.bond_changes = bond_changes
    return reactants, products



import os
import argparse
from typing import List, Union
from rich import print
from tqdm import tqdm
import copy
import pandas as pd
import Bio
import Bio.PDB
from Bio.Data.IUPACData import protein_letters_3to1
import torch
from torch.utils import data
from torch_geometric.data import Data as pygData
from esm import pretrained
from clipzyme.utils.registry import register_object
from clipzyme.utils.wln_processing import get_bond_changes
from clipzyme.utils.screening import process_mapped_reaction
from clipzyme.utils.protein_utils import (
    read_structure_file,
    filter_resolution,
    build_graph,
    compute_graph_edges,
    compute_node_embedding,
)


protein_letters_3to1.update({k.upper(): v for k, v in protein_letters_3to1.items()})


@register_object("reactions_dataset", "dataset")
class ReactionDataset(data.Dataset):
    def __init__(
        self,
        args: argparse.Namespace = None,
        dataset_file_path: str = None,
        esm_dir: str = None,
        protein_cache_dir: str = None,
        use_as_protein_encoder: bool = False,
        use_as_reaction_encoder: bool = False,
    ) -> None:
        """
        Create a dataset of reactions and proteins from a CSV file

        Parameters
        ----------
        args: argparse.Namespace
            Arguments from command line
        dataset_file_path: str
            Path to CSV file with headers ['reaction', 'sequence', 'protein_id', 'cif']
        esm_dir: str
            Path to ESM model directory
        protein_cache_dir: str
            Directory to save/load protein graphs
        use_as_protein_encoder: bool
            Use dataset as protein encoder and do not consider reactions in data filtering
        use_as_reaction_encoder: bool
            Use dataset as reaction encoder and do not consider proteins in data filtering

        Raises
        ------
        ValueError
            If CSV file does not have headers ['reaction', 'sequence', 'protein_id', 'cif']
        """
        super(ReactionDataset, self).__init__()

        if args is None:
            csv_path = dataset_file_path
            esm_dir = esm_dir
            protein_cache_dir = protein_cache_dir
            self.use_as_protein_encoder = use_as_protein_encoder
            self.use_as_reaction_encoder = use_as_reaction_encoder
        else:
            csv_path = args.dataset_file_path
            esm_dir = args.esm_dir
            protein_cache_dir = args.protein_cache_dir
            self.use_as_protein_encoder = args.use_as_protein_encoder
            self.use_as_reaction_encoder = args.use_as_reaction_encoder

        # check if csv file has correct headers
        with open(csv_path, "r") as f:
            headers = f.readline().strip().split(",")
            if (
                "reaction" not in headers
                or "protein_id" not in headers
                or "sequence" not in headers
                or "cif" not in headers
            ):
                raise ValueError(
                    "CSV file must have headers 'reaction', 'sequence', 'protein_id', and 'cif'"
                )

        print("Loading ESM model")
        esm_path = os.path.join(esm_dir, "esm2_t33_650M_UR50D.pt")
        model, alphabet = pretrained.load_model_and_alphabet(esm_path)
        self.esm_model = model
        self.alphabet = alphabet
        self.batch_converter = alphabet.get_batch_converter()

        print("Preparing dataset")
        csv_dataset = pd.read_csv(csv_path)
        csv_dataset = csv_dataset.fillna("")
        self.dataset = self.create_dataset(csv_dataset)

        self.protein_cache_dir = protein_cache_dir

        print(self.SUMMARY_STATEMENT)

    def create_dataset(self, csv_dataset: pd.DataFrame) -> List[dict]:
        """
        Create dataset of reactions and proteins from CSV file

        Parameters
        ----------
        csv_path : str
            Path to CSV file with headers ['reaction', 'sequence', 'cif']

        Returns
        -------
        List[dict]
            _description_
        """

        dataset = []

        for rowid, row in tqdm(
            csv_dataset.iterrows(),
            desc="Building dataset",
            total=len(csv_dataset),
            ncols=100,
        ):
            reactants, products = row["reaction"].split(">>")
            reactants = reactants.split(".")
            products = products.split(".")
            sample = {
                "protein_id": row["protein_id"],
                "sequence": row["sequence"],
                "reaction": row["reaction"],
                "reactants": reactants,
                "products": products,
                "cif_path": row["cif"],
                "sample_id": f"sample_{rowid}",
            }
            # get bond changes
            try:
                sample["bond_changes"] = get_bond_changes(sample["reaction"])
            except Exception as e:
                sample["bond_changes"] = []

            if self.skip_sample(sample):
                continue

            # add reaction sample to dataset
            dataset.append(sample)

        return dataset

    def skip_sample(self, sample: dict) -> bool:
        """
        Skip sample if criteria are not met

        Parameters
        ----------
        sample : dict
            Sample dictionary from dataset

        Returns
        -------
        bool
            True if sample should be skipped
        """
        # if dataset is not used only as reaction encoder
        if not self.use_as_reaction_encoder:
            # if sequence is unknown
            sequence = sample["sequence"]
            if len(sequence) == 0:
                return True

            if len(sequence) > 650:
                return True

            # check if cif file exists
            if not os.path.exists(sample["cif_path"]):
                return True

        # if dataset is not used only as protein encoder
        if not self.use_as_protein_encoder:
            # if no bond changes
            if len(sample["bond_changes"]) == 0:
                return True

        return False

    def create_protein_graph(self, sample: dict) -> Union[pygData, None]:
        """
        Create pyg protein graph from CIF file

        Parameters
        ----------
        sample : dict
            dataset sample

        Returns
        -------
        data
            pygData object with protein graph
        """
        try:
            raw_path = sample["cif_path"]
            sample_id = sample["sample_id"]
            protein_parser = Bio.PDB.MMCIFParser()
            protein_resolution = "residue"
            graph_edge_args = {"knn_size": 10}
            center_protein = True

            # parse pdb
            all_res, all_atom, all_pos = read_structure_file(
                protein_parser, raw_path, sample_id
            )
            # filter resolution of protein (backbone, atomic, etc.)
            atom_names, seq, pos = filter_resolution(
                all_res,
                all_atom,
                all_pos,
                protein_resolution=protein_resolution,
            )
            # generate graph
            data = build_graph(atom_names, seq, pos, sample_id)
            # kNN graph
            data = compute_graph_edges(data, **graph_edge_args)
            if center_protein:
                center = data["receptor"].pos.mean(dim=0, keepdim=True)
                data["receptor"].pos = data["receptor"].pos - center
                data.center = center

            sequence = sample["sequence"]
            data.structure_sequence = sequence

            node_embeddings_args = {
                "model": self.esm_model,
                "model_location": "",
                "alphabet": self.alphabet,
                "batch_converter": self.batch_converter,
            }
            node_embedding = compute_node_embedding(data, **node_embeddings_args)
            # Fix sequence length mismatches
            if len(data["receptor"].seq) != node_embedding.shape[0]:
                print("Computing seq embedding for mismatched seq length")
                AA_seq = ""
                for char in seq:
                    AA_seq += protein_letters_3to1[char]

                data.structure_sequence = AA_seq
                data["receptor"].x = compute_node_embedding(
                    data, **node_embeddings_args
                )
            else:
                data["receptor"].x = node_embedding

            if len(data["receptor"].seq) != data["receptor"].x.shape[0]:
                return None

            if hasattr(data, "x") and not hasattr(data["receptor"], "x"):
                data["receptor"].x = data.x

            if not hasattr(data, "structure_sequence"):
                data.structure_sequence = "".join(
                    [protein_letters_3to1[char] for char in data["receptor"].seq]
                )

            keep_keys = {
                "receptor",
                "structure_sequence",
                ("receptor", "contact", "receptor"),
            }

            data_keys = data.to_dict().keys()
            for d_key in data_keys:
                if d_key not in keep_keys:
                    delattr(data, d_key)

            coors = data["receptor"].pos
            feats = data["receptor"].x
            edge_index = data["receptor", "contact", "receptor"].edge_index
            assert (
                coors.shape[0] == feats.shape[0]
            ), f"Number of nodes do not match between coors ({coors.shape[0]}) and feats ({feats.shape[0]})"

            assert (
                max(edge_index[0]) < coors.shape[0]
                and max(edge_index[1]) < coors.shape[0]
            ), "Edge index contains node indices not present in coors"

            return data

        except Exception as e:
            print(
                f"Could not create protein graph for:  {sample['protein_id']} because of the exception {e}"
            )
            return None

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, index):
        sample = self.dataset[index]
        sample_id = sample["sample_id"]

        try:
            reactants, products = process_mapped_reaction(
                sample["reaction"],
                sample["bond_changes"],
            )
            protein_id = sample["protein_id"]

            item = {
                "reactants": reactants,
                "products": products,
                "protein_id": protein_id,
                "sample_id": sample_id,
                "cif_path": sample["cif_path"],
                "sequence": sample["sequence"],
            }

            if self.protein_cache_dir:
                graph_path_cache = os.path.join(
                    self.protein_cache_dir, f"{protein_id}.pt"
                )
                try:
                    data = torch.load(graph_path_cache)
                    if data is None:
                        data = self.create_protein_graph(item)
                        torch.save(data, graph_path_cache)
                except:
                    data = self.create_protein_graph(item)
                    torch.save(data, graph_path_cache)
            else:
                data = self.create_protein_graph(item)

                item["graph"] = data

            return item

        except Exception as e:
            print(f"Could not load sample {sample_id} because of an exception {e}")

    @property
    def SUMMARY_STATEMENT(self) -> None:
        reactions = [d["reaction"] for d in self.dataset]
        proteins = [d["sequence"] for d in self.dataset]

        statement = f""" 
        DATASET CREATED:
        * Number of samples: {len(self.dataset)}
        * Number of reactions: {len(set(reactions))}
        * Number of proteins: {len(set(proteins))}
        """
        return statement
